// ============================================================================
// Hamster, a free news- and mailserver for personal, family and workgroup use.
// Copyright (c) 1999, Juergen Haible.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
// ============================================================================

unit cFiltersMail;

interface

uses SysUtils, Classes, Global, cArticle, cFiltersBase, cPCRE;

type
  TFilterPatternMail = class( TFilterPatternBase )
    FilterHeader: String;
    ExtractField: boolean;
  end;

Type
  TMailfilterAction = (mfaINVALID, mfaLOAD, mfaKILL, mfaIGNORE, mfaNOTIFY, mfaNOTIFYOFF,
     mfaDEFAULT, mfaSET, mfaADD, mfaDEL, mfaADDACCOUNTS, mfaPOSTTO, mfaADDDEFAULT,
     mfaSetScore, mfaAddScore, mfaLOG, mfaLOGOFF); //PW Mail-KillLog 2003-06-06
Const
  MailfilterAction_Names: Array[TMailfilterAction] of String
     = ('#invalid#', 'load', 'kill', 'ignore', 'notify', 'notifyoff',
         'default', 'set', 'add', 'del', 'addaccounts', 'postto', 'adddefault',
         'setscore', 'addscore', 'log', 'logoff');

Type
  TFilterLineMail = class( TFilterLineBase )
    private
      procedure GetHeaderValue(const TestHdr: String; Mail: TArticle; Out s: String);
      procedure GetHeaderValues(const TestHdr: String; Mail: TArticle; Values: TStrings);
      function SpecialHeader(const TestHdr: String; Mail: TArticle; out Value: String): Boolean;
    protected
      function MatchesAccounts( Mail: TArticle; var ResultAccounts: String ): Boolean;
      function TestHeaderValue( Const HdrVal: String; Mail: TArticle; RE: TPCRE ): Boolean;
    public
      IsFinal      : Boolean; // '='?
      DoAllHeaders : Boolean; // '*'?
      ActionID     : TMailfilterAction;
      ActionPars   : String;
      DefaultField : String;
      Expire       : TDateTime;
      Unless       : Boolean;
      // Next fields had to be set before executing MatchesMailHeaders, used by GetHeaderValue(s)
      SizeOfMsg, Score: Integer;
      //
      function  MatchesMailHeaders( Mail: TArticle; RE: TPCRE; var ResultAccounts: String ): Boolean;
      function  SetFilterLine( Const FileName: String; Const LineNo: Integer;
         Const FilterLine: String ): Boolean; override;
      Function CheckFilterline ( Const Filterline: String ): Boolean;
  end;

  TFiltersMail = class( TFiltersBase )
    public
      TOP_makes_sense  : Boolean;
      Filter_bytes_only: Boolean;
      Procedure Test;
      function  LinesAdd( Const FileName: String; Const LineNo: Integer;
         Const LineText: String ): Integer; override;
      function  IsFilterLine( Const Line: String ): Boolean; override;
      procedure SelectSections( Const SectionIdentifier: String ); override;
      procedure FilterMail(
         Mail: TArticle; Const SizeOfMsg  : Integer;
         DefaultUser      : String;
         Var ResultIgnore, ResultKillIt : Boolean;
         ResultNotifys, ResultKillLogs, ResultUsers, ResultGroups: TStrings;
         Var NotifyReason, AccountReason, PostToReason, Scoring: String;
         Out Score: Integer);
      Procedure Purge;
  end;

implementation

uses uTools, uDateTime, cAccount, cMailrouter, Config, uEncoding,
     cLogFile, cStdForm;

const
  ANY_SENDER        = 'Any-Sender:';
  ALL_SENDERS       = 'All-Senders:';
  ANY_SENDER_MAX    = 6;
  ANY_SENDER_LIST   : array[0..ANY_SENDER_MAX] of String =
                      ( 'From:', 'Apparently-From:', 'Sender:', 'Reply-To:',
                        'X-Sender', 'Envelope-From:', 'X-Envelope-From:' );

  ANY_RECIPIENT     = 'Any-Recipient:';
  ALL_RECIPIENTS    = 'All-Recipients:';
  ANY_RECIPIENT_MAX = 5;
  ANY_RECIPIENT_LIST: array[0..ANY_RECIPIENT_MAX] of String =
                      ( 'To:', 'Apparently-To:', 'CC:', 'BCC:',
                        'Envelope-To:', 'X-Envelope-To:' );

procedure ExpandFilterGroups( HdrName: String; TS: TStrings );
var  i: Integer;
begin
   If HdrName = '' then Exit;
   if HdrName[length(HdrName)]<>':' then HdrName:=HdrName+':';
   if CompareText( ANY_SENDER, HdrName )=0 then begin
      for i:=0 to ANY_SENDER_MAX do TS.Add( ANY_SENDER_LIST[i] );
   end else
   if CompareText( ANY_RECIPIENT, HdrName )=0 then begin
      for i:=0 to ANY_RECIPIENT_MAX do TS.Add( ANY_RECIPIENT_LIST[i] );
   end else
   begin
      TS.Add( HdrName )
   end
end;

// ------------------------------------------------------ TFilterLineMail -----

function TFilterLineMail.MatchesAccounts( Mail: TArticle; var ResultAccounts: String ): Boolean;
var  TestHdrs, MailAddrs: TStringList;
     TestHdr, DefStr, s: String;
     CurrHdr, LastPos, UID, i: Integer;
     LTY: TLocalmailtype; 
     bLocal: Boolean;

     procedure AddMailAddress( TestStr: String );
     var  at: Integer;
     begin
          TestStr := TrimWhSpace( TestStr );
          if length(TestStr)<3 then exit;
          if PosWhSpace( TestStr )>0 then exit;

          if TestStr[1]='<' then System.Delete( TestStr, 1, 1 );
          if TestStr[length(TestStr)]='>' then System.Delete( TestStr, length(TestStr), 1 );

          at := Pos( '@', TestStr );
          if (at<2) or (at=length(TestStr)) then exit;

          MailAddrs.Add( TestStr );
     end;

     procedure AddMailAddresses( HdrVal: String );
     // add all strings which look like mail-addresses to accounts-testlist
     var  i, p, p2: Integer;
          s, addr: String;
     begin
        s := HdrVal;
        // Delete all "xx"-Parts
        p := Pos('"', s);
        While p > 0 do begin
           p2 := 0;
           For i := p+1 to Length(s) do If s[i]='"' then begin
              p2 := i; break
           end;
           If p2>0 then begin
              Delete(s, p, p2-p+1);
              p := Pos('"', s)
           end else begin
              Exit // Kaputte Adresse!
           end
        end;
        // Delete all (xx)-Parts
        p := Pos('(', s);
        While p > 0 do begin
           p2 := Pos(')', s);
           If p2 > p then begin
              Delete(s, p, p2-p+1);
              p := Pos('(', s)
           end else begin
              Exit // Kaputte Adresse!
           end
        end;
        // Separate Addresses
        Repeat
           p := Pos(',', s);
           If p > 0 then begin
              addr := Trim(Copy(s, 1, p-1));
              Delete(s, 1, p)
           end else begin
              addr := Trim(s);
              s := ''
           end;
           p := Pos('<', s);
           p2 := Pos('>', s);
           If (p>0) and (p2>p) then begin
              Addr := Copy(Addr, p+1, p2-p-1)
           end;
           If Addr > '' then AddMailAddress( addr )
        until s = ''
     end;
     
begin
   TestHdrs := TStringList.Create;
   MailAddrs := TStringList.Create;
   try
     ExpandFilterGroups( DefaultField, TestHdrs );
     MailAddrs.Sorted := True;
     MailAddrs.Duplicates := dupIgnore;
     for CurrHdr:=0 to TestHdrs.Count-1 do begin
        TestHdr := TestHdrs[CurrHdr];
        LastPos := -1;
        repeat
           DefStr := Mail.HeaderAfterIdx( TestHdr, LastPos );
           If DefStr > '' then begin
              If DoMimeDecode then DefStr:=DecodeHeadervalue(DefStr);
              AddMailAddresses( DefStr )
           end
        until (DefStr = '') or (not DoAllHeaders);
     end;

     CfgAccounts.Lock;
     try
        for i:=0 to MailAddrs.Count-1 do begin
           if MailAddrs[i]<>'' then begin
              if CfgAccounts.IsLocalMailbox( MailAddrs[i], UID, LTY, bLocal ) then begin
                 s := CfgAccounts.Users.Find(UID).Username;
                 if ResultAccounts<>'' then ResultAccounts:=ResultAccounts+',';
                 ResultAccounts := ResultAccounts + s;
              end else begin
                 Log( LOGID_DEBUG, 'MailFilter.addaccounts: "' + MailAddrs[i] + '" unknown.' );
              end;
           end;
        end
     finally
        CfgAccounts.Unlock
     end;

     if ResultAccounts<>'' then Log( LOGID_DEBUG, 'MailFilter.addaccounts.final: "' + ResultAccounts + '"' );
     Result := ( ResultAccounts <> '' )
   finally
     MailAddrs.Free;
     TestHdrs.Free;
   end
end;

function TFilterLineMail.TestHeaderValue( Const HdrVal: String; Mail: TArticle; RE: TPCRE ): Boolean;
var  Matches, NeedOneOf, HaveOneOf: Boolean;
     PatNo: Integer;
     TestStr: String;
     Pat: TFilterPatternMail;
     Pattern: String;
begin
     Result := True;

     NeedOneOf := False;
     HaveOneOf := False;

     for PatNo:=0 to PatternCount-1 do begin
        Pat := TFilterPatternMail( PatternItem[ PatNo ] );

        If Pat.ExtractField
           then Pattern := Mail[Pat.PatternStr]
           else Pattern := Pat.PatternStr;
        If Pattern = '' then continue;

        if CompareText( Pat.FilterHeader, DefaultField )=0
           then TestStr := HdrVal
           else GetHeaderValue(Pat.FilterHeader,Mail,TestStr);

        if (Pat.SelectType<>' ') or not(HaveOneOf) then begin
           Matches := Pat.Matches(Pattern, TestStr, RE);
           case Pat.SelectType of
              '+': if not Matches then begin Result:=False; break; end;
              '-': if Matches     then begin Result:=False; break; end;
              ' ': begin
                      NeedOneOf := True;
                      if Matches then HaveOneOf:=True;
                   end;
           end;
        end;
     end;

     if NeedOneOf and not HaveOneOf then Result:=False;
end;

Function TFilterLineMail.SpecialHeader(Const TestHdr: String; Mail: TArticle; Out Value: String): Boolean;
Var LastPos, i, z, k: Integer; h, s: String;
begin
   Result := true;
   Value := '';
   If CompareText(TestHdr, 'Bytes:')=0 then Value := IntToStr(SizeOfMsg)
   else If CompareText(TestHdr, 'Score:')=0 then Value := IntToStr(Score)
   else If CompareText(TestHdr, 'Header:')=0 then Value := Mail.FullHeader
   else If CompareText(TestHdr, 'Age:')=0 then begin
      try
          k := Trunc( NowGMT - RfcDateTimeToDateTimeGMT( Mail.Header['Date'], NowGMT ) );
          Value := IntToStr(k)
      except
          Value := '0'
      end
   end
   else If CompareText(TestHdr, 'Top:')=0 then Value := Mail.FullBody
   else If CompareText(TestHdr, 'Toponly:')=0 then begin
      z := Def_NoOfTopLinesToLoad;
      Value := Mail.FullBody;
      For i := 1 to Length(Value)-1 do If (Value[i]=#13) and (Value[i+1]=#10) then begin
         Dec(z);
         If z = 0 then begin
            SetLength(Value, i+1);
            break
         end
      end
   end
   else If CompareText(TestHdr, ALL_RECIPIENTS)=0 then begin
      For i := 0 to ANY_RECIPIENT_MAX do begin
         h := ANY_RECIPIENT_LIST[i];
         LastPos := -1;
         Repeat
            s := Mail.HeaderAfterIdx(h, LastPos);
            If s > '' then begin
               If DoMimeDecode then s:=DecodeHeadervalue(s);
               Value := Value + s + #13#10
            end
         until s=''
      end
   end
   else If CompareText(TestHdr, ALL_SENDERS)=0 then begin
      For i := 0 to ANY_SENDER_MAX do begin
         h := ANY_SENDER_LIST[i];
         LastPos := -1;
         Repeat
            s := Mail.HeaderAfterIdx(h, LastPos);
            If s > '' then begin
               If DoMimeDecode then s:=DecodeHeadervalue(s);
               Value := Value + s + #13#10
            end
         until s=''
      end
   end
   else Result := false
end;

procedure TFilterLineMail.GetHeaderValue(const TestHdr: String; Mail: TArticle; Out s: String);
begin
   If Not SpecialHeader(TestHdr, Mail, s) then s := Mail[TestHdr]
end;

Procedure TFilterLineMail.GetHeaderValues(Const TestHdr: String; Mail: TArticle; Values: TStrings);
Var LastPos: Integer; s: String;
begin
   Values.Clear;
   If SpecialHeader(TestHdr, Mail, s) then Values.Add(s)
   else If Mail.HeaderExists(TestHdr) then begin
      LastPos := -1;
      Repeat
         s := Mail.HeaderAfterIdx(TestHdr, LastPos);
         If s > '' then begin
            If DoMimeDecode then s:=DecodeHeadervalue(s);
            Values.Add(s)
         end
      until (s='') or (Not DoAllHeaders)
   end
end;

function TFilterLineMail.MatchesMailHeaders( Mail: TArticle; RE: TPCRE;
   var ResultAccounts: String ): Boolean;
var  i, CurrHdr: Integer; TestHdrs, HdrValues: TStringList;
begin
   ResultAccounts := '';
   If ActionID = mfaADDACCOUNTS then begin
      // test given header-name[s] for known mail-accounts
      Result := MatchesAccounts( Mail, ResultAccounts )
   end else begin
      // test given patterns
      Result := Defaultfield = '';
      TestHdrs := TStringList.Create;
      HdrValues:= TStringList.Create;
      try
         ExpandFilterGroups( DefaultField, TestHdrs );
         for CurrHdr:=0 to TestHdrs.Count-1 do begin
            GetHeaderValues(TestHdrs[CurrHdr], Mail, HdrValues);
            For i := 0 to HdrValues.Count-1 do begin
               // test given patterns
               Result := TestHeaderValue( HdrValues[i], Mail, RE ); // test patterns
               If Result then break
            end;
            if Result then break
         end;
      finally
         TestHdrs.Free;
         HdrValues.Free;
      end
   end;
   If unless then Result := Not result
end;

Type TVTyp = (vtText, vtField, vtRegex);

function TFilterLineMail.SetFilterLine(Const FileName: String; Const LineNo: Integer; Const FilterLine: String ): Boolean;
Type TFilterMode = (fmStart, fmCommand, fmCommandExt, fmCommandPar,
            fmDefField1, fmDefField2, fmDefField3,
            fmPattern0, fmPattern1, fmPattern2, fmPatternField, fmPatternText,
            fmExpireDate0, fmExpireDate );
Var Mode: TFilterMode;
    Err, s, TmpFilterHeader: String;
    i, p, Stack: Integer;
    done: Boolean;
    c, TmpSelectType, StopChar: Char;
    pat: TFilterPatternMail;

   Procedure InitPattern;
   begin
      Mode := fmPattern0;
      TmpSelectType     := ' ';
      TmpFilterHeader   := '';
      StopChar          := ' ';
      Stack             := 0
   end;

   Procedure CheckExpire(Const StrDate: String);
   begin
      If Expire > 0 then begin
         Err := 'Only one expire-expression allowed in one line'
      end else
      If Length(StrDate)<>8 then begin
         Err := 'Invalid date in expire-expression - format must be YYYYMMDD'
      end else try
         Expire := EncodeDate(StrToInt(Copy(StrDate, 1, 4)),
                              StrToInt(Copy(StrDate, 5, 2)),
                              StrToInt(Copy(StrDate, 7, 2)));
      except
         Err := 'Invalid date "'+s+'" in expire-expression - format must be YYYYMMDD'
      end
   end;

Const Commentchars = ['#', ';'];
      WhiteSpace = [' ',^I];
      ActionChars = ['A'..'Z', 'a'..'z'];
      HeaderChars = [#33..#57, #59..#126];
Var iMFA: TMailfilterAction;
    PatTyp: TPatterntype;
begin
   Err := '';
   Clear;
   fOrgLineNo   := LineNo;
   fOrgFilename := FileName;
   fOrgLineText := Filterline;
   IsFinal      := False;
   ActionID     := mfaINVALID;
   ActionPars   := '';
   DoMimeDecode := False;
   DoAllHeaders := False;
   DefaultField := '';
   Expire       := 0;
   Unless       := false;
   Mode := fmStart;
   p := 0;
   For i := 1 to Length(Filterline) + 1 do begin
      If i <= Length(Filterline) then c := Filterline[i] else c := ' ';
      Case Mode of
         // Analysis begins
         fmStart:
         If c IN Commentchars then break
         else If c IN WhiteSpace then
         else If c IN ActionChars then begin p := i; Mode := fmCommand end
         else If c = '=' then begin
            If IsFinal then Err := 'Doubled final-marker "="'
                       else IsFinal := true;
         end else Err := 'Action expected';
         // Check command Part 1/3
         fmCommand: begin
            done := true;
            If c IN Commentchars then break
            else If c IN WhiteSpace then Mode := fmCommandExt
            else if c IN ActionChars then done := false
            else If c = '(' then Mode := fmCommandPar;
            If done then begin
               s := LowerCase(Copy(Filterline, p, i-p));
               ActionID := mfaINVALID;
               For iMFA := Low(TMailfilteraction) to High(TMailfilteraction) do begin
                  If s=Mailfilteraction_Names[iMFA] then begin
                     ActionID := iMFA;
                     break
                  end
               end;
               If ActionID = mfaInvalid then Err := 'Unknown action "'+s+'"';
               p := i+1
            end
         end;
         fmCommandPar: If c = ')' then begin
            ActionPars := TrimWhSpace(Copy(Filterline, p, i-p));
            Mode := fmDefField1
         end;
         fmCommandExt: Case c of
            '(': begin p := i+1; Mode := fmCommandPar end;
            '~': begin DoMimeDecode := true; Mode:=fmDefField2; p:=i+1 end;
            '*': begin DoAllHeaders := true; Mode:=fmDefField3; p:=i+1 end;
            else begin
               If c IN Commentchars then break
               else If c IN WhiteSpace then
               else If c IN HeaderChars then begin Mode:=fmDefField3; p:=i end
               else Err := 'invalid char "'+c+'" after action.'
            end;
         end;
         // -----------------------------------------------------
         fmDefField1: Case c of
            '~': begin DoMimeDecode := true; Mode:=fmDefField2; p:=i+1 end;
            '*': begin DoAllHeaders := true; Mode:=fmDefField3; p:=i+1 end;
            else begin
               If c IN Commentchars then break
               else If c IN WhiteSpace then
               else If c IN HeaderChars then begin Mode:=fmDefField3; p:=i end
               else Err := 'invalid char "'+c+'" after action.'
            end;
         end;
         fmDefField2: Case c of
            '*': begin DoAllHeaders := true; Mode:=fmDefField3; p:=i+1 end;
            else begin
               If c IN Commentchars then break
               else If c IN WhiteSpace then
               else If c IN HeaderChars then begin Mode:=fmDefField3; p:=i end
               else Err := 'invalid char "'+c+'" after action.'
            end
         end;
         fmDefField3: begin
            done := false;
            Case c of
               ':': done := true;
               else begin
                  If c IN Commentchars then break
                  else If c IN WhiteSpace then done := true
                  else If c IN HeaderChars then
                  else Err := 'invalid char "'+c+'" in def-field'
               end
            end;
            If done then begin
               // Special case Expire
               If (c=':') and (LowerCase(Copy(Filterline, p, i-p))='expire') then begin
                  Mode := fmExpireDate0
               end else
               // Special case Unless
               If (c<>':') and (LowerCase(Copy(Filterline, p, i-p))='unless')
                  and (Not DoAllHeaders) and (Not DoMimeDecode)
               then begin
                  Unless := true;
                  Mode := fmDefField1
               // Default-Field detected!
               end else begin
                  DefaultField := Copy(Filterline, p, i-p)+':';
                  InitPattern
               end
            end
         end;
         // ----------------------------------------
         fmExpireDate0: If c IN['0'..'9'] then begin
            Mode := fmExpireDate;
            p := i
         end else begin
            Err := 'Invalid expire-expression - date must follow directly after "expire:"'
         end;
         fmExpireDate: begin
            done := false;
            If c IN WhiteSpace then done := true
            else If Not (c IN ['0'..'9']) then begin
               Err := 'Invalid char "'+c+'" in expire-expression'
            end;
            If done then begin
               CheckExpire(Copy(Filterline, p, i-p));
               If DefaultField = ''
                  then Mode := fmDefField1
                  else InitPattern
            end
         end;
         // ----------------------------------------
         fmPattern0: Case c of
            '+', '-': begin TmpSelectType := c; Mode := fmPattern1 end;
            '@': begin p := i+1; Mode := fmPatternField end;
            '$', '"', '''': begin p := i+1; Mode := fmPatternText; StopChar := c end;
            '{': begin p := i+1; Mode := fmPatternText; StopChar := '}' end;
            else begin
               If c IN Commentchars then break
               else If c IN WhiteSpace then
               else begin p := i; Mode := fmPatternText end
            end
         end;
         fmPattern1: Case c of
            '@': begin p := i+1; Mode := fmPatternField end;
            '$', '"', '''': begin p := i+1; Mode := fmPatternText; StopChar := c end;
            '{': begin p := i+1; Mode := fmPatternText; StopChar := '}' end;
            else begin
               If c IN WhiteSpace then
               else begin p := i; Mode := fmPatternText end
            end
         end;
         fmPattern2: Case c of
            '$', '"', '''': begin p := i+1; Mode := fmPatternText; StopChar := c end;
            '{': begin p := i+1; Mode := fmPatternText; StopChar := '}' end;
            else begin
               If c IN WhiteSpace then
               else begin p := i; Mode := fmPatternText end
            end
         end;
         fmPatternField: Case c of
            ':': begin
                    TmpFilterHeader := Copy(Filterline, p, i-p+1);
                    Mode := fmPattern2;
                 end;
            else begin
               If c IN HeaderChars then
               else Err := 'invalid char "'+c+'" in special fieldname'
            end
         end;
         fmPatternText: begin
            done := false;
            If (c IN WhiteSpace) and (StopChar = ' ') then done := true
            else If (c=StopChar) and (Stack<=0) then done := true
            else If (c=StopChar) and (Stack<=0) then done := true
            else If StopChar='}' then begin
               If c = '{' then Inc(Stack)
               else If c = '}' then Dec(Stack)
            end;
            If done then begin
               s := Copy(Filterline, p, i-p);
               If (StopChar=' ') and (LowerCase(Copy(s, 1, 7))='expire:')
                  and (TmpSelectType = ' ') and (TmpFilterHeader='')
               then begin
                  CheckExpire(Copy(s, 8, Length(s)-7))
               end else
               If (StopChar = ' ') and (s<>'*') and (Copy(s,1,1)<>'%') then begin
                  Err := 'Missing "" around text-pattern "'+s+'"'
               end else begin
                  PatTyp := ptString;
                  Case StopChar of
                     '}': PatTyp := ptRegExp;
                     '"', '$': ;
                     '''': PatTyp := ptExactString;
                     else begin
                        If s = '*' then PatTyp := ptWildcard
                        else If Copy(s, 1, 1)='%' then PatTyp := ptValue
                        else begin err := 'invalid pattern-type detected'; break end;
                     end
                  end;
                  pat := TFilterPatternMail.Create(TmpSelectType, PatTyp, s);
                  pat.ExtractField := StopChar = '$';
                  If TmpFilterHeader = ''
                     then pat.FilterHeader := DefaultField
                     else pat.FilterHeader := TmpFilterHeader;
                  PatternAdd( pat );
               end;
               InitPattern
            end
         end;
      end;
      If Err>'' then break;
   end;
   If Err = '' then Case Mode of
      fmStart: If IsFinal then Err := 'Missing action after "="';
      fmCommandExt, fmDefField1, fmPattern0: ;
      fmCommand: Err := 'Internal error, parser ended in impossible state #'+IntToStr(Ord(fmCommand));
      fmCommandPar: Err := 'Missing ")" to end action-parameter';
      fmDefField2, fmDefField3: Err := 'Missing default field';
      fmPattern1, fmPattern2, fmPatternText: Err := 'Missing pattern';
      fmPatternField: Err := 'Missing end of pattern-field';
      fmExpireDate0, fmExpireDate: Err := 'Missing end of expire-rule';
      else Err := 'Internal error, no defined reaction for end-state #'+IntToStr(Ord(fmCommand));
   end;
   If (Err = '') then begin
      If ActionID=mfaADDACCOUNTS then begin
         If DefaultField = '' then Err := 'Missing field containing accounts'
         else If fFilterPatterns.Count>0 then Err := 'No pattern allowed for this type of action';
      end else begin
         If (DefaultField > '') and (fFilterPatterns.Count=0) then Err := 'Default-field defined without any pattern'
      end
   end;
   Result := Err = '';
   If Result
      then LastSetError := ''
      else LastSetError := '"'+Copy(Filterline,1,i)+'" => '+Err
end;

// --------------------------------------------------------- TFiltersMail -----

function TFiltersMail.LinesAdd( Const FileName: String; Const LineNo: Integer; Const LineText: String ): Integer;
var  lin: TFilterLineMail;
begin
     lin := TFilterLineMail.Create;
     if lin.SetFilterLine( FileName, LineNo, LineText ) then begin
        Result := fFilterLines.Add( lin );
     end else begin
        Log( LOGID_WARN, 'Filter-line ignored: ' + LineText );
        Log( LOGID_WARN, 'Reason: ' + lin.LastSetError );
        lin.Free;
        Result := -1;
     end;
end;

function TFiltersMail.IsFilterLine( Const Line: String ): Boolean;
begin
   Result := (Line>'') and (Line[1] in ['=','a'..'z','A'..'Z'])
end;

procedure TFiltersMail.SelectSections( Const SectionIdentifier: String );
var  LineNo: Integer;
     Line  : TFilterLineMail;
begin
   inherited SelectSections( SectionIdentifier );
   TOP_makes_sense := False;
   for LineNo:=0 to LinesCount-1 do begin
      Line := TFilterLineMail( LinesItem[ LineNo ] );
      if Line.ActionID in [mfaKILL, mfaIGNORE] then begin
         TOP_makes_sense := True;
         break;
      end;
   end;
end;

procedure TFiltersMail.FilterMail(
   Mail: TArticle; Const SizeOfMsg  : Integer;
   DefaultUser      : String;
   Var ResultIgnore, ResultKillIt : Boolean;
   ResultNotifys, ResultKillLogs, ResultUsers, ResultGroups: TStrings;
   Var NotifyReason, AccountReason, PostToReason, Scoring: String;
   Out Score: Integer);

   Function ParseDest(s : string; var Usr, Folder, Flags : string) : Boolean;
   var i : integer;
   begin
     s := StringReplace(s, '/', '\', [rfReplaceAll, rfIgnoreCase]);
     i := pos(':', s);  //Parse Flags out
     if (i>0) then begin
        Flags := copy(s, i+1, length(s));
        s := trim(copy(s, 1, i-1));
     end else begin
        Flags := ''
     end;

     i := pos('\', s);  //Parse Folders out
     if (i>0) then begin
        Folder := copy(s, i+1, length(s));
        s := trim(copy(s, 1, i-1));
     end else begin
        Folder := ''
     end;

     Usr    := trim(s); //Rest is user-part
     Result := (Usr<>'')
   end;

   Procedure ClearResultUsers;
   Var i: integer;
       IMAP_MTObj: PMailToObj;
   begin
      {HSR} {IMAP-Folder 03}
      For i := 0 to ResultUsers.count-1 do begin
         IMAP_MTObj := pMailToObj(ResultUsers.Objects[i]);
         If Assigned(IMAP_MTObj) then Dispose(IMAP_MTObj)
      end;
      {/HSR}
      ResultUsers.Clear
   end;

var  LineNo, i, k  : Integer;
     Line          : TFilterLineMail;
     Parser        : TParser;
     ResAccounts, s: String;
     WantNotify    : Boolean;
     WantKillLog   : Boolean; //PW Mail-KillLog 2003-06-06
     IMAP_MTObj    : pMailToObj;
     Usr, Folder, Flags : String;
begin
   ResultIgnore := False;
   ResultKillIt := False;
   WantNotify   := False;
   ResultNotifys.Clear;
{PW}{Mail-KillLog 2003-06-06}
   WantKillLog  := False;
   ResultKillLogs.Clear;
{/PW}{Mail-KillLog 2003-06-06}
   ClearResultUsers;
   ResultGroups.Clear;
   NotifyReason := '';
   AccountReason := '';
   PostToReason := '';
   Scoring := '';
   Score := 0;
   Parser := TParser.Create;
   try
     if DefaultUser='' then DefaultUser:=Def_Postmaster;
     for LineNo:=0 to LinesCount-1 do begin
        Line := TFilterLineMail( LinesItem[ LineNo ] );
        Line.SizeOfMsg := SizeOfMsg;
        Line.Score := Score;
        if Line.MatchesMailHeaders( Mail, RegexFilter, ResAccounts ) then begin
           Log( LOGID_DEBUG, 'MailFilter (+): ' + Line.AsString );
           case Line.ActionID of
              // undo a previous kill/ignore
              mfaLOAD: begin
                 ResultIgnore := False;
                 ResultKillIt := False;
                 NotifyReason := '';
              end;
              // don't load, but delete
              mfaKILL: begin
                 ResultKillIt := True;
                 NotifyReason := NotifyReason + '-> ' + Line.AsString + #13#10;
              end;
              // don't load, don't delete
              mfaIGNORE: begin
                 ResultKillIt := False;
                 ResultIgnore := True;
                 NotifyReason := NotifyReason + '-> ' + Line.AsString + #13#10;
              end;
              // notify if "killed" or "ignored"
              mfaNOTIFY: begin
                 WantNotify := True;
                 ResultNotifys.Clear;
                 Parser.Parse( Line.ActionPars, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    if s<>'' then ResultNotifys.Add( s );
                 end;
              end;
              // do NOT notify anyone
              mfaNOTIFYOFF: begin   // JAWO 21.09.01 (Notify none)
                 WantNotify := false;
                 ResultNotifys.Clear;
              end;
              // new default-user
              mfaDEFAULT: begin
                 DefaultUser := Line.ActionPars;
              end;
              // set/start new recipient-list
              mfaSET: begin
                 AccountReason := AccountReason + #13#10+' '+Line.AsString;
                 ClearResultUsers;
                 Parser.Parse( Line.ActionPars, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    if ParseDest(s, Usr, Folder, Flags) then begin
                      new(IMAP_MTObj);
                      IMAP_MTObj^.IMAPFolder := Folder;
                      IMAP_MTObj^.IMAPFlags  := Flags;
                      ResultUsers.AddObject( Usr, Pointer(IMAP_MTObj));
                    end
                 end;
              end;
              // add to recipient-list
              mfaADD: begin
                 AccountReason := AccountReason + #13#10+' '+Line.AsString;
                 Parser.Parse( Line.ActionPars, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    if ParseDest(s, Usr, Folder, Flags) then begin
                       if (ResultUsers.IndexOf( Usr ) < 0) then begin
                          new(IMAP_MTObj);
                          IMAP_MTObj^.IMAPFolder := Folder;
                          IMAP_MTObj^.IMAPFlags  := Flags;
                          ResultUsers.AddObject( Usr, tObject(IMAP_MTObj))
                       end
                    end
                 end
              end;
              // delete from recipient-list
              mfaDEL: begin
                 AccountReason := AccountReason + #13#10+' '+Line.AsString;
                 Parser.Parse( Line.ActionPars, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    ParseDest(s, Usr, Folder, Flags);
                    k := ResultUsers.IndexOf( Usr );
                    if k>=0 then begin
                       if assigned(pMailToObj(ResultUsers.Objects[k])) then
                         dispose(pMailToObj(ResultUsers.Objects[k]));
                       ResultUsers.Delete( k )
                    end;
                 end;
              end;
              // add matching accounts to recipient-list
              mfaADDACCOUNTS: begin
                 AccountReason := AccountReason + #13#10+' '+Line.AsString;
                 Parser.Parse( ResAccounts, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    if ParseDest(s, Usr, Folder, Flags) then begin
                       new(IMAP_MTObj);
                       IMAP_MTObj^.IMAPFolder := Folder;
                       IMAP_MTObj^.IMAPFlags  := Flags;
                       ResultUsers.AddObject( Usr, tObject(IMAP_MTObj));
                    end
                 end
              end;
              // add matching accounts to recipient-list
              mfaADDDEFAULT: begin
                 AccountReason := AccountReason + #13#10+' '+Line.AsString;
                 ResultUsers.Add( DefaultUser  );
              end;
              // post to newsgroups
              mfaPOSTTO: begin
                 PostToReason := Line.AsString;
                 Parser.Parse( Line.ActionPars, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    if s<>'' then ResultGroups.Add( s );
                 end;
              end;
              // Scoring
              mfaSETSCORE: begin
                 Score := StrToInt(Line.ActionPars);
                 Scoring := Scoring + #13#10+' '+Line.AsString
              end;
              mfaADDSCORE: begin
                 Score := Score + StrToInt(Line.ActionPars);
                 Scoring := Scoring + #13#10+' '+Line.AsString
              end;
{PW}{Mail-KillLog 2003-06-06}
              // log if "killed" or "ignored"
              mfaLOG: begin
                 WantKillLog := True;
                 ResultKillLogs.Clear;
                 Parser.Parse( Line.ActionPars, ',' );
                 for i:=0 to Parser.Count-1 do begin
                    s := TrimWhSpace( Parser.sPart( i, '' ) );
                    if s<>'' then ResultKillLogs.Add( s );
                 end;
              end;
              // do NOT log
              mfaLOGOFF: begin   // JAWO 21.09.01 (Notify none)
                 WantKillLog := false;
                 ResultKillLogs.Clear;
              end;
{/PW}
           end;
           if Line.IsFinal then break; // line preceded with "="
        end else begin
           {Log( LOGID_FULL, 'MailFilter (-): ' + Line.AsString );}
        end
     end;
     // use default-user, if no recipients/newsgroups were selected
     if (ResultUsers.Count=0) and (ResultGroups.Count=0) then begin
        if DefaultUser<>'' then begin
           Parser.Parse( DefaultUser, ',' );
           for i:=0 to Parser.Count-1 do begin
              s := TrimWhSpace( Parser.sPart( i, '' ) );
              if s<>'' then ResultUsers.Add( s );
           end;
        end;
     end;

     // fill list of notification-recipients if none were given with "notify()"
     if WantNotify and (ResultIgnore or ResultKillIt) then begin
        // send notifications to determined recipients ...
        if ResultNotifys.Count=0 then ResultNotifys.Text:=ResultUsers.Text;
        // .. or finally to "postmaster"
        if ResultNotifys.Count=0 then ResultNotifys.Add( Def_Postmaster );
     end;

     {PW}{Mail-KillLog 2003-06-06}
     // fill list of log-recipients if none were given with "log()"
     if WantKillLog and (ResultIgnore or ResultKillIt) then begin
        // send notifications to determined recipients ...
        if ResultKillLogs.Count=0 then ResultKillLogs.Text:=ResultUsers.Text;
        // .. or finally to "postmaster"
        if ResultKillLogs.Count=0 then ResultKillLogs.Add( Def_MailKillLogRecipient );
     end;
     {/PW}
   finally
     Parser.Free
   end
end;

procedure TFiltersMail.Purge;
Var ok: Boolean; LineNo: Integer; LineText: String;
begin
   ok := false;
   for LineNo:=fFilterFile.Count-1 downto 0 do begin
      LineText := TrimWhSpace( fFilterFile[LineNo] );
      if LineText > '' then If IsFilterLine( LineText ) then begin
         With TFilterLineMail.Create do try
            If SetFilterLine( '', 0, LineText ) then begin
               If (Expire > 0) and (Expire <= Now) then begin
                  If Def_Mail_ExpiredFilterentries_Delete then begin
                     fFilterFile.Delete (LineNo);
                     LogFile.Add ( LOGID_INFO, TrGlF(kLog, 'Scorefile.ExpiredEntry.Deleted',
                        'Expired Scorefile-Entry deleted: "%s"', LineText))
                  end else begin
                     fFilterFile.RawLines[LineNo] := '# Expired! # '+fFilterFile.RawLines[LineNo];
                     LogFile.Add ( LOGID_INFO, TrGlF(kLog, 'Scorefile.ExpiredEntry.commented',
                        'Expired Scorefile-Entry commented out: "%s"', LineText))
                  end;
                  ok := true
               end
            end
         finally Free end
      end
   end;
   If ok then Save //HSR //INCLUDE
   //fFilterFile.SaveToFile ( fFilterFilename );
end;

procedure TFiltersMail.Test;
var  i, Counter: Integer;
     Line  : TFilterLineMail;
     LineText, s: String;
     bUseLog: Boolean;
begin
   If fFilterFile.Count = 0 then Exit;
   Log( LOGID_SYSTEM, 'Testing mail-filter-file' );
   Line := TFilterLineMail.Create;
   Counter := 0;
   bUseLog := false;
   try
      With fFilterFile do begin
         For i:=0 to Count-1 do begin
            LineText := TrimWhSpace( Lines[i] );
            if IsFilterLine( LineText ) then begin
               If Line.SetFilterLine( FileName[i], LineNr[i], LineText ) then begin
                  If Line.ActionID = mfaLOG then bUseLog := true;
                  s := LowerCase(Line.DefaultField);
                  If (Copy(s, 1, 4)='any-') or (Copy(s, 1, 4)='any_') then begin
                     If (s<>LowerCase(ANY_SENDER)) and (s<>LowerCase(ANY_RECIPIENT)) then begin
                        Log( LOGID_WARN, 'Warning for Mail-filter-file'
                          + IIf(FileName[i]>'', ' "'+FileName[i]+'"')
                          + ', Line '+IntToStr(LineNr[i])+': ' + LineText );
                        Log( LOGID_WARN, 'Warning: Unknown default-field "'+Line.defaultfield+'" - do you mean "'
                          +ANY_SENDER+'" or "'+ANY_RECIPIENT+'" instead?')
                     end
                  end else
                  If (Copy(s, 1, 4)='all-') or (Copy(s, 1, 4)='all_') then begin
                     If (s<>LowerCase(ALL_SENDERS)) and (s<>LowerCase(ALL_RECIPIENTS)) then begin
                        Log( LOGID_WARN, 'Warning for Mail-filter-file'
                          + IIf(FileName[i]>'', ' "'+FileName[i]+'"')
                          + ', Line '+IntToStr(LineNr[i])+': ' + LineText );
                        Log( LOGID_WARN, 'Warning: Unknown default-field "'+Line.defaultfield+'" - do you mean "'
                          +ALL_SENDERS+'" or "'+ALL_RECIPIENTS+'" instead?')
                     end
                  end;
               end else begin
                  Log( LOGID_WARN, 'Error in Mail-filter-file "'+ FileName[i] + '"' ); //HSR //INCLUDE (Chg));
                  Log( LOGID_WARN, 'Error in line '+IntToStr(LineNr[i])+': ' + LineText ); //HSR //INCLUDE (Chg));
                  Log( LOGID_WARN, 'Error description: ' + line.LastSetError );
                  Inc(Counter)
               end
            end else
            If (LineText>'') and (Not (LineText[1] IN['[', ';', '#'])) then begin
               Log( LOGID_WARN, 'Error in Mail-filter-file "'+ FileName[i] + '"' ); //HSR //INCLUDE (Chg));
               Log( LOGID_WARN, 'Error in line '+IntToStr(LineNr[i])+': ' + LineText );
               Log( LOGID_WARN, 'Error description: No comment, no section and no action');
               Inc(Counter)
            end
         end
      end
   finally
      Line.free
   end;
   If Counter = 0
      then Log( LOGID_INFO, 'Mail-filter-file: Test OK')
      else Log( LOGID_SYSTEM, 'Mail-filter-file: '+IntToStr(Counter)+' errors found');
   If bUseLog then Def_AlwaysUseUIDL := true
end;

function TFilterLineMail.CheckFilterline(
  const Filterline: String): Boolean;
Var i: Integer;
begin
   Result := SetFilterline('', 0, Filterline);
   If Result then begin
      Log ( LOGID_INFO, 'Filterline: '+filterline );
      Log ( LOGID_INFO, 'Check ok, interpretation of filterline follows' );
      If Expire > 0 then Log ( LOGID_INFO, '- Expire-date of line: '+DateToStr(Expire) );
      Log ( LOGID_INFO, '- Action: "'+MailfilterAction_Names[ActionID] + '('+ActionPars+')"' );
      If DoAllheaders then Log ( LOGID_INFO, '- Check multiple headerfields (*)' );
      If IsFinal then Log ( LOGID_INFO, '- Stop with this line if matches (=)');
      If DefaultField > '' then Log ( LOGID_INFO, '- Default header-field (if missing the line is ignored): "'+DefaultField+'"' );
      If Unless then Log ( LOGID_INFO, '- Unless used = line matches, if patterns don''t match' );
      For i := 0 to PatternCount-1 do begin
         With TFilterPatternMail( PatternItem[ i ] ) do begin
            Log ( LOGID_INFO, '- Pattern '+IntToStr(i+1)+'/'+IntToStr(PatternCount));
            Case SelectType of
               '+': Log ( LOGID_INFO, '  - Pattern must match (+)');
               '-': Log ( LOGID_INFO, '  - Pattern mustn''t match (-)');
               else Log ( LOGID_INFO, '  - One of these kind of pattern must match');
            end;
            Log ( LOGID_INFO, '  - Filter on header "'+FilterHeader+'"');
            Case PatternType of
               ptWildcard: Log ( LOGID_INFO, '  - Matches always: *');
               ptRegExp: Log ( LOGID_INFO, '  - Matches, if regular expression matches on header: {'+PatternStr+'}');
               ptString: If ExtractField
                  then Log ( LOGID_INFO, '  - Matches, if header contains the value of header "'+PatternStr+'"')
                  else Log ( LOGID_INFO, '  - Matches, if header contains (case-insensitive): "'+PatternStr+'"');
               ptExactString: Log ( LOGID_INFO, '  - Matches, if header contains (case-sensitive): '''+PatternStr+'''');
               ptValue: Log ( LOGID_INFO, '  - Matches, if numeric value of header '+PatternStr+' '+IntToStr(PatternInt));
               else raise Exception.Create('Internal error - TFilterLineMail.CheckFilterline: Unknown PatternType');
            end
         end
      end;
      Log ( LOGID_INFO, 'End of interpretation.' );
   end else begin
      Log ( LOGID_INFO, 'Check failed - '+LastSetError );
   end
end;

end.
